# Databricks notebook source
# https://danijar.com/variable-sequence-lengths-in-tensorflow/  : Variable length sequences to RNN

# COMMAND ----------

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import glob
import json
from datetime import datetime
import itertools
import random

#tf imports
import tensorflow as tf
from tensorflow.nn import rnn_cell
from tensorflow.saved_model import tag_constants

assert np.__version__ == "1.15.4" and pd.__version__ == "0.23.4", "np: {}, pd: {}".format(np.__version__, pd.__version__)

# COMMAND ----------

# MAGIC %run ./func

# COMMAND ----------

log_dir, model_dir = set_dirs('new_CV')

# COMMAND ----------

df_model_a = get_model_a_data()
df_model_a.head()

# COMMAND ----------

input_df = df_model_a.loc[:, [segment, 'current', 'sea_temp', 'sampling_int']]

data = input_df.values
intervals = df_model_a.loc[:, 'sampling_int'].values

print(data.shape, intervals.shape)
print(input_df.count())
input_df.head()

# COMMAND ----------

# Set up tf graph
learning_rate = 0.01
n_hidden = 64
dim_input = data.shape[1]
n_input_steps = 50
n_output_steps = 1
pred_steps = 1

def model_AD_graph():
  tf.reset_default_graph()

  assert n_input_steps >= pred_steps, 'input steps must be > pred steps'

  # Graph input
  x = tf.placeholder("float", [None, n_input_steps, dim_input], name='x')
  t = tf.placeholder("float", [None, pred_steps, n_output_steps], name='t')
  y = tf.placeholder("float", [None, pred_steps ,n_output_steps], name='y')

  # Define weights
  weights = {
      'out': tf.Variable(tf.random_normal([n_hidden+1,n_output_steps]))  # +1 because of concatenated time until prediction input
  }
  biases = {
      'out': tf.Variable(tf.zeros([n_output_steps]))
  }

  cell = rnn_cell.LSTMCell(n_hidden, name='LSTM_cell')

  outputs, _ = tf.nn.dynamic_rnn(cell, x, dtype=tf.float32)  
  preds = tf.matmul(tf.reshape(tf.concat([outputs[:, -pred_steps:, :], t], axis=2), (-1, (n_hidden+1))), weights['out']) + biases['out']
  preds = tf.reshape(preds, (-1, pred_steps, n_output_steps), name='preds')

  # Define loss and optimizer
  loss_op = tf.losses.mean_squared_error(labels=y, predictions=preds)
  optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate)
  train_op = optimizer.minimize(loss_op)

  init = tf.global_variables_initializer()
  saver = tf.train.Saver()
  
  return train_op, preds, init, x, t, y

train_op, preds, init, x, t, y = model_AD_graph()

# COMMAND ----------

fold_ranges = get_blocked_fold_ranges()

# COMMAND ----------

def print_fold_ranges():
  plt.close()
  fig, (ax1,ax2, ax3, ax4) = plt.subplots(4,1, figsize=(20,7.5), sharex=True)
  ax1.plot(input_df[segment], color='black')
  ax1.set_ylabel('DTS temperature')
  #print(ax1.y_label)
  ax2.plot(input_df['current'], color='black')
  ax2.set_ylabel('Electric current')
  ax3.plot(input_df['sea_temp'], color='black')
  ax3.set_ylabel('Sea temperature')
  ax4.plot(intervals, color='black')
  ax4.set_ylabel('Interval length')
  ax4.set_xlabel('Date')
  colors = ['red', 'blue', 'yellow', 'green', 'orange']
  for i, block_ranges in enumerate(fold_ranges):
    for r in block_ranges:
      for a in [ax1, ax2, ax3, ax4]:
        a.axvspan(r.start, r.stop, alpha=0.5, color=colors[i])
  ticks = [i for i in range(0, len(df_model_a), 7000)]
  ticks.append(len(df_model_a)-1)
  labels = [pd.to_datetime(df_model_a.iloc[i,:]['timestamp'], unit='s').date() for i in range(0, len(df_model_a), 7000)]
  labels.append(pd.to_datetime(df_model_a.iloc[-1,:]['timestamp'], unit='s').date())
  plt.xticks(ticks, labels)
  display(plt.show())
  print(labels)
print_fold_ranges()

# COMMAND ----------

class Model_A_data:  
  def __init__(self, test_fold, val_fold, fold_ranges, n_input_steps=50, pred_steps=1):
    assert test_fold != val_fold, 'test fold cannot be equal to val fold'
    self.test_fold = test_fold
    self.val_fold = val_fold
    self.fold_ranges = fold_ranges
    self.n_input_steps = n_input_steps
    self.pred_steps = pred_steps
    train_d = np.concatenate([input_df.iloc[fold_ranges[fold][block]] for fold in range(n_folds) for block in range(n_blocks) if fold != val_fold and fold != test_fold])
    interval_d = np.concatenate([intervals[fold_ranges[fold][block]] for fold in range(n_folds) for block in range(n_blocks) if fold != val_fold and fold != test_fold])
    self.inputs_means = np.mean(train_d, axis=0)
    self.inputs_stds = np.std(train_d, axis=0)
    self.targets_mean = np.mean(train_d[:,0], axis=0)
    self.targets_std = np.std(train_d[:,0], axis=0)
    self.intervals_mean = np.mean(interval_d)
    self.intervals_std = np.std(interval_d)
  
  def get_inputs_multiple_folds(self, folds):
    temp_inputs = []
    temp_targets = []
    temp_intervals = []
    for fold in folds:
      fold_inputs, fold_targets, fold_intervals = self.__get_fold_inputs(fold)
      temp_inputs.append(fold_inputs)
      temp_targets.append(fold_targets)
      temp_intervals.append(fold_intervals)
    return np.concatenate(temp_inputs), np.concatenate(temp_targets), np.concatenate(temp_intervals)
  
  def get_inputs_single_fold(self, fold):
    assert fold == self.test_fold or fold == self.val_fold, 'should be val or test fold when getting inputs for single fold'
    return(self.__get_fold_inputs(fold))
    
  def __get_fold_inputs(self, fold):
    temp_inputs = []
    temp_targets = []
    temp_intervals = []
    for br in fold_ranges[fold]:
      block_inputs, block_targets, block_intervals = self.__get_inputs_single_block(br)
      temp_inputs.append(block_inputs)
      temp_targets.append(block_targets)
      temp_intervals.append(block_intervals)
    fold_inputs, fold_targets, fold_intervals= np.concatenate(temp_inputs) , np.concatenate(temp_targets), np.concatenate(temp_intervals)
    fold_inputs = (fold_inputs - self.inputs_means) / self.inputs_stds
    fold_targets = (fold_targets - self.targets_mean) / self.targets_std
    fold_intervals = (fold_intervals - self.intervals_mean) / self.intervals_std
    return fold_inputs, fold_targets, fold_intervals
  
  def __get_inputs_single_block(self, block_range):
    block_inputs, block_targets, block_intervals = get_inputs_targets_intervals(data[block_range,:], intervals[block_range],  self.n_input_steps, self.pred_steps)
    return block_inputs, block_targets, block_intervals
  
  def undo_normalize(self, temp_preds):
    return (temp_preds * self.targets_std) + self.targets_mean

# COMMAND ----------

is_testing = True
n_epochs = 1000
batch_size = 128
initial_patience = 50
n_folds = 5

MSEs_train = []
MSEs_val = []
MSEs_test = []
MSEs_train_norm = []
MSEs_val_norm = []

file_suffix = 'Model_D'
log_filename = log_dir + file_suffix + '.txt' if file_suffix else None
model_filename = None
if log_filename: assert not os.path.isfile(log_filename), "path " + log_filename + " is already in use"
if not file_suffix:
  print("Training model without saving or logging")
else:
  write_to_log('writing log to {} and models to {}'.format(log_filename, model_filename), log_filename)

args = {'lr': learning_rate,
        'n hidden': str(n_hidden),
        'dim input':dim_input,
        'input steps':n_input_steps,
        'output steps':n_output_steps,
        'pred steps':pred_steps,
        'n folds': n_folds,
        'n epochs': n_epochs,
        'batch size': batch_size,
        'patience': initial_patience,
        'data shape': data.shape,
        'fold ranges': fold_ranges}
write_to_log(str(args), log_filename)
  
for test_fold in range(n_folds):
  if not is_testing:
    test_fold = -1
  for val_fold in range(n_folds):
    if test_fold != val_fold or not is_testing:
      write_to_log('test fold {} val fold {}'.format(test_fold, val_fold), log_filename)
      
      model_a_data = Model_A_data(test_fold, val_fold, fold_ranges)
      train_inputs, train_targets, train_intervals = model_a_data.get_inputs_multiple_folds([i for i in range(n_folds) if i !=val_fold and i != test_fold])
      val_inputs, val_targets, val_intervals = model_a_data.get_inputs_single_fold(val_fold)
      if is_testing:
        test_inputs, test_targets, test_intervals =  model_a_data.get_inputs_single_fold(test_fold)
      
      epoch_indicies = [i for i in range(len(train_inputs))]
      min_val_loss, min_train_loss = np.inf, np.inf
      patience = initial_patience
      
      with tf.Session() as sess:
        sess.run(init)
        for epoch in range(n_epochs):
          np.random.shuffle(epoch_indicies)
          epoch_inputs, epoch_targets, epoch_t = train_inputs[epoch_indicies], train_targets[epoch_indicies], train_intervals[epoch_indicies]
          for batch in range(0, len(train_inputs), batch_size):
            batch_x, batch_y, batch_t  = epoch_inputs[batch:batch+batch_size], epoch_targets[batch:batch+batch_size], epoch_t[batch:batch+batch_size]
            sess.run(train_op, feed_dict={x:batch_x, t:batch_t, y:batch_y})
          epoch_val_loss = MSE(sess.run(preds, feed_dict ={x:val_inputs, t:val_intervals})[:,-1,:], val_targets[:,-1,:], model=model_a_data)
          if epoch_val_loss < min_val_loss and patience:
            min_val_loss = epoch_val_loss
            val_preds_epoch = sess.run(preds, feed_dict={x:val_inputs, t:val_intervals})
            train_preds_epoch = sess.run(preds, feed_dict={x:train_inputs, t:train_intervals})
            train_MSE_epoch = MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:], model=model_a_data)
            test_MSE = MSE(sess.run(preds, feed_dict={x:test_inputs, t:test_intervals})[:,-1,:], test_targets[:,-1,:], model=model_a_data) if is_testing else -1
            write_to_log("\tEpoch {} finished. For 1-step preds training loss is {}, validation loss is {} and test loss is {}".format(epoch, train_MSE_epoch, min_val_loss, test_MSE), log_filename)
            patience = initial_patience
          elif patience:
            patience -= 1
          else:
            write_to_log("Stopped training at epoch {} due to patience running out".format(epoch), log_filename)
            break
      MSE_train = MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:], model=model_a_data)
      MSE_val = MSE(val_preds_epoch[:,-1,:], val_targets[:,-1,:], model=model_a_data)
      MSEs_train.append(MSE_train)
      MSEs_val.append(MSE_val)
      MSEs_test.append(test_MSE)
      MSEs_train_norm.append(MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:]))
      MSEs_val_norm.append(MSE(val_preds_epoch[:,-1,:], val_targets[:,-1,:]))
      write_to_log("Accuracy for one step preds train:{} val:{}\n".format(MSE_train, MSE_val), log_filename)
  if not is_testing:
    break
    
write_to_log('average MSEs across all folds train:{} val:{} test:{}'.format(np.mean(MSEs_train), np.mean(MSEs_val), np.mean(MSEs_test)), log_filename)
write_to_log('MSE std across all folds train:{} val:{} test:{}'.format(np.std(MSEs_train), np.std(MSEs_val), np.std(MSEs_test)), log_filename)
if not is_testing:
  write_to_log('MSE val for each fold: {} & {} & {} & {} & {}'.format(MSEs_val[0],MSEs_val[1],MSEs_val[2],MSEs_val[3],MSEs_val[4]), log_filename)
else:
  s = ''
  for m in MSEs_test:
    s += '& {} '.format(m)
  write_to_log(s, log_filename)

# COMMAND ----------




# COMMAND ----------


