# Databricks notebook source
# https://danijar.com/variable-sequence-lengths-in-tensorflow/  : Variable length sequences to RNN

# COMMAND ----------

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import glob
import json
from datetime import datetime
import itertools

from sklearn.linear_model import LinearRegression

assert np.__version__ == "1.15.4" and pd.__version__ == "0.23.4", "np: {}, pd: {}".format(np.__version__, pd.__version__)

# COMMAND ----------

# MAGIC %run ./func

# COMMAND ----------

class Model_A_data:  
  def __init__(self, test_fold, val_fold, fold_ranges, n_input_steps=50, pred_steps=1):
    assert test_fold != val_fold, 'test fold cannot be equal to val fold'
    self.test_fold = test_fold
    self.val_fold = val_fold
    self.fold_ranges = fold_ranges
    self.n_input_steps = n_input_steps
    self.pred_steps = pred_steps
    train_d = np.concatenate([input_df.iloc[fold_ranges[fold][block]] for fold in range(n_folds) for block in range(n_blocks) if fold != val_fold and fold != test_fold])
    interval_d = np.concatenate([intervals[fold_ranges[fold][block]] for fold in range(n_folds) for block in range(n_blocks) if fold != val_fold and fold != test_fold])
    self.inputs_means = np.mean(train_d, axis=0)
    self.inputs_stds = np.std(train_d, axis=0)
    self.targets_mean = np.mean(train_d[:,0], axis=0)
    self.targets_std = np.std(train_d[:,0], axis=0)
    self.intervals_mean = np.mean(interval_d)
    self.intervals_std = np.std(interval_d)
  
  def get_inputs_multiple_folds(self, folds):
    temp_inputs = []
    temp_targets = []
    temp_intervals = []
    for fold in folds:
      fold_inputs, fold_targets, fold_intervals = self.__get_fold_inputs(fold)
      temp_inputs.append(fold_inputs)
      temp_targets.append(fold_targets)
      temp_intervals.append(fold_intervals)
    return np.concatenate(temp_inputs), np.concatenate(temp_targets), np.concatenate(temp_intervals)
  
  def get_inputs_single_fold(self, fold):
    assert fold == self.test_fold or fold == self.val_fold, 'should be val or test fold when getting inputs for single fold'
    return(self.__get_fold_inputs(fold))
    
  def __get_fold_inputs(self, fold):
    temp_inputs = []
    temp_targets = []
    temp_intervals = []
    for br in fold_ranges[fold]:
      block_inputs, block_targets, block_intervals = self.__get_inputs_single_block(br)
      temp_inputs.append(block_inputs)
      temp_targets.append(block_targets)
      temp_intervals.append(block_intervals)
    fold_inputs, fold_targets, fold_intervals= np.concatenate(temp_inputs) , np.concatenate(temp_targets), np.concatenate(temp_intervals)
    fold_inputs = (fold_inputs - self.inputs_means) / self.inputs_stds
    fold_targets = (fold_targets - self.targets_mean) / self.targets_std
    fold_intervals = (fold_intervals - self.intervals_mean) / self.intervals_std
    return fold_inputs, fold_targets, fold_intervals
  
  def __get_inputs_single_block(self, block_range):
    block_inputs, block_targets, block_intervals = get_inputs_targets_intervals(data[block_range,:], intervals[block_range],  self.n_input_steps, self.pred_steps)
    return block_inputs, block_targets, block_intervals
  
  def undo_normalize(self, temp_preds):
    return (temp_preds * self.targets_std) + self.targets_mean

# COMMAND ----------

fold_ranges = get_blocked_fold_ranges()

data_df = get_model_a_data()
input_df = data_df.loc[:, [segment, 'current', 'sea_temp', 'sampling_int']]
input_df.count()

data = input_df.values
intervals = input_df.loc[:, 'sampling_int'].values

# COMMAND ----------

is_testing = True
n_epochs = 1000
batch_size = 128
initial_patience = 50
n_input_steps = 50
n_dims = 4

MSEs_train = []
MSEs_val = []
MSEs_test = []
MSEs_train_norm = []
MSEs_val_norm = []
MSEs_test_all = []


MSEs_val_int = []
MSEs_train_int = []

MSEs_test_int = []
MSEs_train_int = []
MSEs_val_int = []

all_MSEs_test_int = []
all_MSEs_val_int = []
all_MSEs_train_int = []

for test_fold in range(n_folds):
  if not is_testing:
    test_fold = -1
  for val_fold in range(n_folds):
    if test_fold != val_fold or not is_testing:

      model_a_data = Model_A_data(test_fold, val_fold, fold_ranges)
      train_inputs, train_targets, train_intervals = model_a_data.get_inputs_multiple_folds([i for i in range(n_folds) if i !=val_fold and i != test_fold])
      val_inputs, val_targets, val_intervals = model_a_data.get_inputs_single_fold(val_fold)
      if is_testing:
        test_inputs, test_targets, test_intervals =  model_a_data.get_inputs_single_fold(test_fold)
      
      reg_with_interval = LinearRegression().fit(np.concatenate([train_inputs.reshape(-1, n_input_steps*n_dims),
                                                                 train_intervals.reshape(-1,1)], axis=1), train_targets.flatten())

      # Get predictions on val data.
      val_preds_int = reg_with_interval.predict(np.concatenate([val_inputs.reshape(-1, n_input_steps*n_dims), val_intervals.reshape(-1,1)], axis=1))
      train_preds_int = reg_with_interval.predict(np.concatenate([train_inputs.reshape(-1, n_input_steps*n_dims), train_intervals.reshape(-1,1)], axis=1))
      test_preds_int = reg_with_interval.predict(np.concatenate([test_inputs.reshape(-1, n_input_steps*n_dims), test_intervals.reshape(-1,1)], axis=1))

      # Undo normalization, calculate MSE

      MSEs_val_int.append(MSE(val_preds_int, val_targets.flatten(), model=model_a_data))
      MSEs_train_int.append(MSE(train_preds_int, train_targets.flatten(), model=model_a_data))
      MSEs_test_int.append(MSE(test_preds_int, test_targets.flatten(), model=model_a_data))
      print('t{}v{}: train: {}, val: {}, test: {}'.format(test_fold, val_fold, MSEs_train_int[-1], MSEs_val_int[-1], MSEs_test_int[-1]))
  if not is_testing:
    break
  else:
    all_MSEs_test_int.append(MSEs_test_int)
    all_MSEs_val_int.append(MSEs_val_int)
    all_MSEs_train_int.append(MSEs_train_int)
    MSEs_test_int = []
    MSEs_train_int = []
    MSEs_val_int = []

    
print("With interval:")
print('Test & {} & {} & {} & {} & {}'.format(np.mean(all_MSEs_test_int[0]),np.mean(all_MSEs_test_int[1]),np.mean(all_MSEs_test_int[2]),np.mean(all_MSEs_test_int[3]),np.mean(all_MSEs_test_int[4])))
print('Val & {} & {} & {} & {} & {}'.format(np.mean(all_MSEs_val_int[0]),np.mean(all_MSEs_val_int[1]),np.mean(all_MSEs_val_int[2]),np.mean(all_MSEs_val_int[3]),np.mean(all_MSEs_val_int[4])))
print('Train & {} & {} & {} & {} & {}'.format(np.mean(all_MSEs_train_int[0]),np.mean(all_MSEs_train_int[1]),np.mean(all_MSEs_train_int[2]),np.mean(all_MSEs_train_int[3]),np.mean(all_MSEs_train_int[4])))

# COMMAND ----------

print('LR & {} & {} & {}'.format(np.mean(all_MSEs_train_int), np.mean(all_MSEs_val_int), np.mean(all_MSEs_test_int)))

# COMMAND ----------

for i, t in enumerate(all_MSEs_test_int):
  s = str(i+1)
  
  t_copy = t.copy()
  t_copy.insert(i, 0)
  
  for ind in t_copy:
    s += ' & {}'.format(ind if ind else '\\-')
  print(s + '\\\ \hline')

# COMMAND ----------

np.array(all_MSEs_test_int).flatten()

# COMMAND ----------

all_MSEs_test_int

# COMMAND ----------


