# Databricks notebook source
from scipy.stats import describe
from sklearn.metrics import roc_curve, auc, precision_recall_curve

import matplotlib.pyplot as plt
import json

# COMMAND ----------

fault_dir = '/dbfs/mnt/renewablesstudents/user-write/faults_allDataApril/'

# COMMAND ----------

def get_normal_data(metadata_path, column):
  with open(metadata_path, 'r') as file:
    metadata = json.load(file)
  time_range = {'start_timestamp': pd.Timestamp(metadata['start']), 'end_timestamp':pd.Timestamp(metadata['end'])}
  if column=='random':
    return df_by_time_range(df_dts[metadata['segments'] + ['new_timestamp', 'timestamp']], time_range).drop(['new_timestamp'], axis=1).copy()
  else:
    #if fault_type in ['typeI', 'typeIII']:
      #return df_by_time_range(df_segment, time_range).drop(['new_timestamp', 'sampling_interval'], axis=1).copy()
    return df_by_time_range(df_dts[metadata['segments'] + ['new_timestamp', 'timestamp']], time_range).drop(['new_timestamp'], axis=1).copy()
    

# COMMAND ----------

def get_result_summary_V2(result_list, normal_result, metadata):
  assert metadata['type'] == 'instant' or metadata['type'] == 'drifting'
  columns = metadata['columns']
  begin_step = metadata['begin_step'] if metadata['type'] == 'instant' else 0
  located_anomaly = [result[0] in columns and result[0]>-1  for result in result_list[begin_step:]]
  TP = np.sum(located_anomaly)
  FN = len(located_anomaly) - TP
  located_normal = [result[0] == -1 for result in normal_result]
  TN = np.sum(located_normal)
  FP = len(located_normal) - TN
  return {"TP":TP, "FN":FN, "TN":TN, "FP":FP, "anomaly_detected":TP>0  , "healty_detected":FP>0, "1st_found":located_anomaly[0]}

def calculate_precision_recall(TP, FN, FP):
  precision = TP / (TP + FP)
  recall = TP / (TP + FN)
  return precision, recall

def calculate_F_beta(precision, recall, beta=0.1):
  return (1+beta**2)*precision*recall / (beta**2*precision + recall)


def get_AD_results_V2(anomalyDetector, fault_dir, log=False, column='random', folds=[0,1,2,3,4], types=['typeI', 'typeII', 'typeIII', 'typeIV']):
  if not log:
    log = Log(False, False)
  results = []
  log.write('Finding anomaly detection results for column {}'.format(column))
  log.write('Faults are loaded from {}'.format(fault_dir))
  for fold in folds:
    fold_results = {}
    log.write('fold: {}'.format(fold))
    
    fold_dir = fault_dir + 'test_fold{}/'.format(fold) + 'column_{}/'.format(column)  
    for fault_type in types:
      fold_results[fault_type] = {}
      type_results = []
      data_paths = glob.glob(fold_dir + fault_type + '/*.csv')
      metadata_paths = glob.glob(fold_dir + fault_type + '/*.json')
      assert len(data_paths) > 0, 'data paths is empty: {}'.format(fold_dir + fault_type + '/*.csv')
      for index, data_path in enumerate(data_paths):
        metadata_path = metadata_paths[index]
        fault_df, metadata = get_fault_data(data_path, metadata_path)
        normal_df = get_normal_data(metadata_path, column)
        assert set(list(normal_df)).issubset(set(list(fault_df))), 'fault df {} should contain all {} in normal df'.format(list(fault_df), list(normal_df))
        assert len(fault_df) == len(normal_df), 'length of {} should equal length of {}'.format(len(fault_df), len(normal_df))
        type_results.append(get_result_summary_V2(anomalyDetector.is_anomalous(fault_df, metadata, fold), 
                                                  anomalyDetector.is_anomalous(normal_df, metadata, fold),
                                                  metadata))
        
      fold_results[fault_type]['all'] = type_results
      fold_results[fault_type]['aggregated'] = {'TP': np.sum([ts['TP'] for ts in type_results]),
                                               'FN': np.sum([ts['FN'] for ts in type_results]),
                                               'TN': np.sum([ts['TN'] for ts in type_results]),
                                               'FP': np.sum([ts['FP'] for ts in type_results]),
                                               "anomaly_detected": np.mean([ts['anomaly_detected'] for ts in type_results]),
                                               "healty_detected": np.mean([ts['healty_detected'] for ts in type_results]),
                                               "1st_found": np.mean([ts['1st_found'] for ts in type_results]),
                                               }
      
      precision, recall = calculate_precision_recall(fold_results[fault_type]['aggregated']['TP'],
                                                     fold_results[fault_type]['aggregated']['FN'],
                                                     fold_results[fault_type]['aggregated']['FP'])
      fold_results[fault_type]['precision'] = precision
      fold_results[fault_type]['recall'] = recall
      F_beta = calculate_F_beta(precision, recall)
      fold_results[fault_type]['F'] = F_beta
      log.write('\t' + fault_type + ': ' + str(fold_results[fault_type]['aggregated']) + ' (precision: {}, recall: {}, F: {})'.format(precision, recall, F_beta))
      results.append(fold_results)

  #log.write('Aggregated results:')
  
  #all_faults = {ft:[] for ft in types}
  #for fold in folds:
   # for fault_type in types:
      #all_faults[fault_type].append( (results[fold][fault_type]['precision'],
                                      #results[fold][fault_type]['recall'],
                                      #results[fold][fault_type]['F'],
                                      #fold_results[fault_type]['aggregated']["anomaly_detected"],
                                      #fold_results[fault_type]['aggregated']["healty_detected"],
                                      #fold_results[fault_type]['aggregated']["1st_found"],
                                     #))
  #for fault_type, fault_list in all_faults.items():
    #log.write(fault_type + ': ' + 'avg precision: {}, avg recall: {}, avg F: {} \n\t proportion anomalies detected: {}, proportion healthy wrongly detected: {}, proportion of anomalies found immediately: {}'.format(np.mean([ts[0] for ts in fault_list]),
    #np.mean([ts[1] for ts in fault_list]), np.mean([ts[2] for ts in fault_list]), np.mean([ts[3] for ts in fault_list]), np.mean([ts[4] for ts in fault_list]), np.mean([ts[5] for ts in fault_list])))
                                       
  return results

# COMMAND ----------

def _get_anomaly_scores(anomalyDetector, fault_dir,  column, fold, fault_type, log=False):
  if not log:
    log = Log(False, False)
  log.write('Finding anomaly scores for column {}. fold {}. fault type {}'.format(column, fold, fault_type))
  log.write('Faults are loaded from {}'.format(fault_dir))
  assert fault_type in ['typeI', 'typeII', 'typeIII', 'typeIV'], 'fault type' + str(fault_type) + ' not valid'    
  
  fold_dir = fault_dir + 'test_fold{}/'.format(fold) + 'column_{}/'.format(column)  
  data_paths = glob.glob(fold_dir + fault_type + '/*.csv')#[1:2]
  metadata_paths = glob.glob(fold_dir + fault_type + '/*.json')#[1:2]
  assert len(data_paths) > 0, 'data paths is empty: {}'.format(fold_dir + fault_type + '/*.csv')
  fault_scores = []
  norm_scores = []
  for index, data_path in enumerate(data_paths):
    #if index == 30:
    #  print('WARNING: ONLY FAULT 30 IS USED in _get_anomaly_scores in func_AD_evaluation')
    metadata_path = metadata_paths[index]
    fault_df, metadata = get_fault_data(data_path, metadata_path)
    normal_df = get_normal_data(metadata_path, column)
    assert set(list(normal_df)).issubset(set(list(fault_df))), 'fault df {} should contain all {} in normal df'.format(list(fault_df), list(normal_df))
    assert len(fault_df) == len(normal_df), 'length of {} should equal length of {}'.format(len(fault_df), len(normal_df))
    fault_scores.append(anomalyDetector.get_score(fault_df, metadata, fold))
    norm_scores.append(anomalyDetector.get_score(normal_df, metadata, fold))
  return fault_scores, norm_scores

# COMMAND ----------

def get_all_anomaly_scores(anomalyDetector, models_dir, fault_dir, columns=[6290, 17712, 32028], folds=[0,1,2,3,4], fault_types=['typeI', 'typeII', 'typeIII', 'typeIV'], log=False, save_path=False):
  if save_path:
    with open(save_path, 'a+') as file:
      file.seek(0,0)
      assert not file.read(), 'file is not empty'
  auc_scores = {f:{c:{} for c in columns} for f in fault_types}
  false_positive_rates = {f:{c:{} for c in columns} for f in fault_types}
  true_positive_rate = {f:{c:{} for c in columns} for f in fault_types}
  thresholds = {f:{c:{} for c in columns} for f in fault_types}
  precisions = {f:{c:{} for c in columns} for f in fault_types}
  recalls = {f:{c:{} for c in columns} for f in fault_types}
  pr_thresholds = {f:{c:{} for c in columns} for f in fault_types}
  score_list_faults = {f:{c:{} for c in columns} for f in fault_types}
  score_list_normal = {f:{c:{} for c in columns} for f in fault_types}
  
  if not log:
    log = Log(False, False)
  for fault_type in fault_types:  
    for column in columns:
      global df_segment
      global segment
      global df_joined
      if column != 'random':
        df_segment, segment  = load_segment_data(subset=SUBSET, column=column)
        df_joined = join_temp_curr_seatemp(df_segment, df_current, df_seatemp)
        print('segment', segment)
      if 'CAE' in models_dir or 'spatiotemporal_model' in models_dir:
        ad = anomalyDetector(models_dir)
      else:
        #print('else')
        ad = anomalyDetector(models_dir.format(column,'{}','{}'))
      #print('md', ad.model_dir, models_dir)
      for fold in folds:
        f_scores, n_scores = _get_anomaly_scores(ad, fault_dir,  column, fold, fault_type, log=log)
        labels = [1 for _ in f_scores] + [0 for _ in n_scores]
        scores =  f_scores + n_scores 
        fpr, tpr, th = roc_curve(labels, scores)
        pr,re,tr = precision_recall_curve(labels, scores)
        auc_scores[fault_type][column][fold] = auc(fpr, tpr)
        false_positive_rates[fault_type][column][fold] = fpr.tolist()
        true_positive_rate[fault_type][column][fold] = tpr.tolist()
        thresholds[fault_type][column][fold] = th.tolist()
        precisions[fault_type][column][fold] = pr.tolist()
        recalls[fault_type][column][fold] = re.tolist()
        pr_thresholds[fault_type][column][fold] = tr.tolist()
        score_list_faults[fault_type][column][fold] = f_scores
        score_list_normal[fault_type][column][fold] = n_scores
  
  out_dict = {'auc_scores':auc_scores,
              'false_positive_rates':false_positive_rates,
              'true_positive_rates':true_positive_rate,
              'thresholds':thresholds,
              'recalls':recalls,
              'precisions':precisions,
              'pr_threshold': pr_thresholds,
              'score_list_faults':score_list_faults,
              'score_list_normal': score_list_normal
             }
  if save_path:
    with open(save_path, 'a') as file:
      json.dump(out_dict, file)
      log.write('results written to ' + save_path)
  return json.dumps(out_dict)

# COMMAND ----------

class LSTM_AE_AnomalyDetector():
  """
  spatial AD model.
  """
  def __init__(self, model_dir, quantile=0.999, threshold_multiplier=1):
    self.model_dir = model_dir
    
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.preds_error_mean = None
    self.lu_piv = None
    self.anomaly_threshold = None
    self.params_path = None
    self.quantile=quantile
    self.threshold_multiplier = threshold_multiplier

  def _is_anomaly_from_prediction_error(self, preds_error):
    anomaly_likelihood_score = self.__pred_error_likelihood(preds_error)
    #print(anomaly_likelihood_score)
    if anomaly_likelihood_score > self.anomaly_threshold:
      return column, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def __pred_error_likelihood(self, pred_errors):
    tmp = (pred_errors - self.preds_error_mean).T    
    return tmp.T @ lu_solve(self.lu_piv, tmp)
  
  def set_mean_var_anomalyThreshold(self):
    val_inputs = self.data_by_fold.get_validation_data()
    val_pred_errors = get_rec_error_from_saved_params(self.params_path, val_inputs).squeeze()
    self.preds_error_mean = val_pred_errors.mean(axis=0)
    
    print(val_inputs.shape, val_pred_errors.shape)
    
    preds_error_cov = np.cov(val_pred_errors, rowvar=False)
    self.lu_piv = lu_factor(preds_error_cov)
    scores = np.zeros(len(val_pred_errors))
    for i,val_error in enumerate(val_pred_errors):
      scores[i] = self.__pred_error_likelihood(val_error)
    self.anomaly_threshold = self.threshold_multiplier * np.quantile(scores, self.quantile)
    #self.anomaly_threshold = np.quantiles(scores, 0.999) #scores.mean() + 3 * scores.std()
    print('anomaly threshold set to:', self.anomaly_threshold)

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = Temporal_Autoencoder_data_by_fold(self.test_fold, self.val_fold, fold_ranges)
      self.params_path = self.model_dir.format(test_fold, val_fold)
      self.set_mean_var_anomalyThreshold()
      
  def get_score(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    df_joined = join_temp_curr_seatemp(data, df_current, df_seatemp)
    inputs = self.data_by_fold.get_inputs(df_joined)
    pred_errors = get_rec_error_from_saved_params(self.params_path, inputs).squeeze()
    scores = []
    for pred_error in pred_errors:
      scores.append(self.__pred_error_likelihood(pred_error))
    return np.max(scores)
    
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    df_joined = join_temp_curr_seatemp(data, df_current, df_seatemp)
    inputs = self.data_by_fold.get_inputs(df_joined)
    pred_errors = get_rec_error_from_saved_params(self.params_path, inputs).squeeze()
    is_anomaly = [(-1, -1) for _ in range(WINDOW_SIZE-1)]
    for pred_error in pred_errors:
      is_anomaly.append(self._is_anomaly_from_prediction_error(pred_error))
    assert len(is_anomaly) == len(data), 'is_anomaly should be same length as input dataframe. Check INPUT_STEPS ({}) and OUTPUT_STEPS({})'.format(INPUT_STEPS, OUTPUT_STEPS)
    return is_anomaly

# COMMAND ----------

def rec_error_likelihood(recs, mean, var):
  temp = recs - mean
  norm_arr = np.square(temp)/var
  #return np.max(norm_arr)/np.mean(norm_arr), np.argmax(norm_arr)
  return np.sum(norm_arr), np.argmax(norm_arr)

class SpatialAnomalyDetector():
  """
  spatial AD model.
  """
  def __init__(self, model_dir):
    self.model_dir = model_dir
    
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.rec_error_mean = None
    self.rec_error_var = None
    self.anomaly_threshold = None
    
    self.params_path = None

  def _is_anomaly_from_reconstruction_error(self, recon_error):
    anomaly_likelihood_score, anomaly_location = rec_error_likelihood(recon_error, self.rec_error_mean, self.rec_error_var)
    #print(anomaly_likelihood_score)
    if anomaly_likelihood_score > self.anomaly_threshold:
      return anomaly_location, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def set_mean_var_anomalyThreshold(self):
    val_data = self.data_by_fold.get_validation_data()
    val_rec_errors = get_rec_error_from_saved_params(self.params_path, val_data)
    self.rec_error_mean = val_rec_errors.mean(axis=0)
    self.rec_error_var = val_rec_errors.var(axis=0)
    scores = np.zeros(len(val_rec_errors))
    for i,val_error in enumerate(val_rec_errors):
      scores[i] = rec_error_likelihood(val_error, self.rec_error_mean, self.rec_error_var)[0]
    self.anomaly_threshold = scores.mean() + 3 * scores.std()
    print('anomaly threshold set to:', self.anomaly_threshold)

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = Spatial_data_by_fold(self.test_fold, self.val_fold, fold_ranges)
      self.params_path = self.model_dir.format(test_fold, val_fold)
      self.set_mean_var_anomalyThreshold()
  
  def get_score(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    df_orig = df_by_time_range(df_dts, tr).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    recs = get_rec_error_from_saved_params(self.params_path, self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values))
    scores = []
    for rec_error in recs:
      scores.append(rec_error_likelihood(rec_error, self.rec_error_mean, self.rec_error_var)[0])
    return np.max(scores)
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    df_orig = df_by_time_range(df_dts, tr).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    
    recs = get_rec_error_from_saved_params(self.params_path, self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values))
    is_anomaly = []
    for rec in recs:
      is_anomaly.append(self._is_anomaly_from_reconstruction_error(rec))
    #print(is_anomaly[100:])
    return is_anomaly

# COMMAND ----------

def get_pred_error_from_saved_params(params_path, inputs, targets, intervals):
   with tf.Session() as sess:
    saver.restore(sess, params_path)
    return targets - get_predictions(sess, inputs, intervals)

class LSTMPredictionsAnomalyDetector():
  """
  spatial AD model.
  """
  def __init__(self, model_dir, dbf=False, quantile=0.999, th_multiplier=1):
    if not dbf:
      dbf = Temporal_prediction_data_by_fold
    self.model_dir = model_dir
    
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.preds_error_mean = None
    self.lu_piv = None
    self.anomaly_threshold = None
    
    self.params_path = None
    self.dbf = dbf
    
    self.th_multiplier = th_multiplier
    self.quantile = quantile

  def _is_anomaly_from_prediction_error(self, preds_error):
    anomaly_likelihood_score = self.__pred_error_likelihood(preds_error)
    if anomaly_likelihood_score > self.anomaly_threshold:
      return column, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def __pred_error_likelihood(self, pred_errors):
    tmp = (pred_errors - self.preds_error_mean).T    
    return tmp.T @ lu_solve(self.lu_piv, tmp)
  
  def set_mean_var_anomalyThreshold(self):
    val_inputs, val_targets, val_intervals = self.data_by_fold.get_validation_data()
    val_pred_errors = get_pred_error_from_saved_params(self.params_path, val_inputs, val_targets, val_intervals)
    val_pred_errors = val_pred_errors.reshape((len(val_pred_errors), -1))
    self.preds_error_mean = val_pred_errors.mean(axis=0)
    preds_error_cov = np.atleast_2d(np.cov(val_pred_errors, rowvar=False))
    self.lu_piv = lu_factor(preds_error_cov)
    scores = np.zeros(len(val_pred_errors))
    for i,val_error in enumerate(val_pred_errors):
      scores[i] = self.__pred_error_likelihood(val_error)
    #print(describe(scores))
    #quantiles = 
    #print('quantiles', quantiles, )
    #plt.close()
    #plt.boxplot(scores)
    self.anomaly_threshold = self.th_multiplier * np.quantile(scores, self.quantile)#scores.mean() + 3 * scores.std()
    print('anomaly threshold set to:', self.anomaly_threshold)

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = self.dbf(self.test_fold, self.val_fold, fold_ranges)
      self.params_path = self.model_dir.format(test_fold, val_fold)
      self.set_mean_var_anomalyThreshold()
  
  def get_score(self,data,metadata,test_fold):
    self._update_test_fold_if_changed(test_fold)
    input_dataframe = join_temp_curr_seatemp(data, df_current, df_seatemp)
    inputs, targets, intervals = self.data_by_fold.get_inputs_targets_intervals_from_df(input_dataframe)
    pred_errors = get_pred_error_from_saved_params(self.params_path, inputs, targets, intervals)
    pred_errors = pred_errors.reshape((len(pred_errors), -1))
    scores = []
    for pred_error in pred_errors:
      scores.append(self.__pred_error_likelihood(pred_error))
    return np.max(scores)
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    input_dataframe = join_temp_curr_seatemp(data, df_current, df_seatemp)
    inputs, targets, intervals = self.data_by_fold.get_inputs_targets_intervals_from_df(input_dataframe)
    pred_errors = get_pred_error_from_saved_params(self.params_path, inputs, targets, intervals)
    pred_errors = pred_errors.reshape((len(pred_errors), -1))
    is_anomaly = [(-1, -1) for _ in range(INPUT_STEPS+OUTPUT_STEPS-1)]
    for pred_error in pred_errors:
      is_anomaly.append(self._is_anomaly_from_prediction_error(pred_error))
    assert len(is_anomaly) == len(input_dataframe), 'is_anomaly should be same length as input dataframe. Check INPUT_STEPS ({}) and OUTPUT_STEPS({})'.format(INPUT_STEPS, OUTPUT_STEPS)
    return is_anomaly

# COMMAND ----------

class LSTM_PCA_PredictionsAnomalyDetector():
  """
  spatiotemproal AD model using LSTM to predict next PCA decomposition of dts values.
  """
  def __init__(self, model_dir, quantile=0.999, dbf=False):
    if not dbf:
      dbf = Spatiotemporal_data_by_fold_PCA
    self.model_dir = model_dir
    
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.preds_error_mean = None
    self.preds_error_var = None
    self.anomaly_threshold = None
    
    self.params_path = None
    self.dbf = dbf
    self.quantile = quantile

  def _is_anomaly_from_prediction_error(self, preds_error):
    anomaly_likelihood_score, location = self.__error_likelihood(preds_error)
    if anomaly_likelihood_score > self.anomaly_threshold:
      return location, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def __error_likelihood(self, pred_errors):
    temp = pred_errors - self.preds_error_mean
    norm_arr = np.square(temp)/self.preds_error_var
    return np.max(norm_arr)/np.mean(norm_arr), np.argmax(norm_arr)
  
  def _get_PCA_prediction_loss_from_params(self, PCA_inputs, DTS_targets, intervals):
    with tf.Session() as sess:
      saver.restore(sess, self.params_path)
      predictions = get_predictions(sess, PCA_inputs, intervals)
      return DTS_targets - self.data_by_fold.get_sequence_from_pca(predictions)
  
  def set_mean_var_anomalyThreshold(self):
    val_inputs, _, val_intervals = self.data_by_fold.get_validation_data()
    _, val_DTS_targets, _ = self.data_by_fold.get_validation_data_not_PCA()
    val_pred_errors = self._get_PCA_prediction_loss_from_params(val_inputs, val_DTS_targets, val_intervals)
    val_pred_errors = val_pred_errors.reshape((len(val_pred_errors), -1))
    print('val_pred_errors', val_pred_errors.shape)
    self.preds_error_mean = val_pred_errors.mean(axis=0)
    self.preds_error_var = val_pred_errors.var(axis=0)
    scores = np.zeros(len(val_pred_errors))
    for i,val_error in enumerate(val_pred_errors):
      scores[i] = self.__error_likelihood(val_error)[0]
    self.anomaly_threshold = np.quantile(scores, self.quantile)
    print('anomaly threshold set to:', self.anomaly_threshold)

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = self.dbf(self.test_fold, self.val_fold, fold_ranges)
      self.params_path = self.model_dir.format(test_fold, val_fold)
      self.set_mean_var_anomalyThreshold()
      
  def get_score(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    df_orig = df_by_time_range(df_dts_joined, tr).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    inputs_PCA, targets_DTS, intervals = self.data_by_fold.get_inputs_targets_intervals_from_df(df_orig)
    pred_errors = self._get_PCA_prediction_loss_from_params(inputs_PCA, targets_DTS, intervals) #get_pred_error_from_saved_params(self.params_path, inputs, targets, intervals)
    pred_errors = pred_errors.reshape((len(pred_errors), -1))
    scores = []
    for pred_error in pred_errors:
      scores.append(self.__error_likelihood(pred_error)[0])
    return np.max(scores)
      
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    df_orig = df_by_time_range(df_dts_joined, tr).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    inputs_PCA, targets_DTS, intervals = self.data_by_fold.get_inputs_targets_intervals_from_df(df_orig)
    pred_errors = self._get_PCA_prediction_loss_from_params(inputs_PCA, targets_DTS, intervals) #get_pred_error_from_saved_params(self.params_path, inputs, targets, intervals)
    pred_errors = pred_errors.reshape((len(pred_errors), -1))
    is_anomaly = [(-1, -1) for _ in range(INPUT_STEPS+OUTPUT_STEPS-1)]
    for pred_error in pred_errors:
      is_anomaly.append(self._is_anomaly_from_prediction_error(pred_error))
    assert len(is_anomaly) == len(df_orig), 'is_anomaly should be same length as input dataframe. Check INPUT_STEPS ({}) and OUTPUT_STEPS({})'.format(INPUT_STEPS, OUTPUT_STEPS)
    return is_anomaly
