# Databricks notebook source
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import glob
import json
import random
import shutil
from datetime import datetime

assert np.__version__ == "1.15.4" and pd.__version__ == "0.23.4", "np: {}, pd: {}".format(np.__version__, pd.__version__)

pd.set_option('mode.chained_assignment', "raise")

# COMMAND ----------

n_segments = 39516
n_folds = 5
batch_size = 128

# COMMAND ----------

base_url = '/dbfs/mnt/renewablesstudents/user-write/'
def load_segment_data(column=None, subset=False):
  if not column:
    df_tmp = pd.read_csv(base_url + 'circuit_1/part1tmp.csv')
    assert subset==False, "subset is only supported when column in set"
  else:
    df_tmp = pd.read_csv(base_url + 'circuit_1/part2_{}.csv'.format(column))
  df_tmp['new_timestamp'] = pd.to_datetime(df_tmp.timestamp, unit='s')
  df_tmp['sampling_interval'] = df_tmp.timestamp - df_tmp.timestamp.shift(1)
  df_tmp.loc[0, 'sampling_interval'] = 0
  #df_tmp = df_tmp.iloc[1:,:]
  if not column:
    df_tmp = df_tmp[df_tmp.new_timestamp <= pd.to_datetime('2019 2 1')]
  elif subset:
    df_tmp = df_tmp[df_tmp.new_timestamp <= pd.to_datetime('2018 2 16')]
  else:
    df_tmp = df_tmp[df_tmp.new_timestamp <= pd.to_datetime('2019 4 3')]
  df_tmp = df_tmp.reset_index(drop=True)
  segment = '34327.421' if not column else list(df_tmp)[0]
  return df_tmp, segment

def load_DTS_data(subset = False):
  path = glob.glob(base_url + 'DTS_new_format_april2019/*.parquet')[0] if subset else base_url + 'DTS_new_format_april2019'
  df_tmp = pd.read_parquet(path)
  df_tmp['new_timestamp'] = pd.to_datetime(df_tmp.timestamp, unit='s')
  df_tmp['sampling_interval'] = df_tmp.timestamp - df_tmp.timestamp.shift(1)
  df_tmp = df_tmp.iloc[1:,:]
  df_tmp = df_tmp[(df_tmp.new_timestamp <= pd.to_datetime('2019 4 3'))]
  df_tmp = df_tmp.reset_index(drop=True)
  return df_tmp

def df_from_json(filepath):
  with open(filepath) as file:
    js = json.load(file)
  return pd.DataFrame(js['timeSeriesList'][0]['timeSeries'])

def load_current_data(subset = False):
  print("Warning: Current is not resampled!")
  df_current = df_from_json('/dbfs/mnt/renewablesstudents/dow-dts-v3/bazefield_data/90636_DOW-WFO-OS-132kV-30E01+R01 GT1-CurrentL3_AVG.json')[['t', 'v']]
  df_current['t'] = df_current['t'] *1e-3
  df_current['new_timestamp'] = pd.to_datetime(df_current.t, unit='s')
  if subset:
    df_current = df_current[(df_current.new_timestamp <=  pd.to_datetime('2018 2 16')) & (df_current.new_timestamp >=  pd.to_datetime('2017 11 1'))]
  else:
    df_current = df_current[(df_current.new_timestamp <=  pd.to_datetime('2019 4 3')) & (df_current.new_timestamp >=  pd.to_datetime('2017 11 1'))]
  #df_current = df_current[(df_current.t + 3000>= df_tmp.timestamp.iloc[0]) & (df_current.t <= df_tmp.timestamp.iloc[-1])]
  #df_current = df_current.set_index('new_timestamp').resample('10T').mean().reset_index()
  #df_current['v'] = df_current['v'].interpolate()
  #df_current['t'] = df_current['t'].interpolate()
  return df_current.reset_index(drop=True)

def load_seatemp_data(subset = False):
  #df_seatemp = pd.read_json(base_url + 'from_bz/40406_DOW-EFS-WaterTemp.json')[['t', 'v']] #/
  df_seatemp = df_from_json('/dbfs/mnt/renewablesstudents/dow-dts-v3/bazefield_data/40406_DOW-EFS-WaterTemp.json')[['t', 'v']]
  df_seatemp['t'] = df_seatemp['t'] *1e-3
  df_seatemp['new_timestamp'] = pd.to_datetime(df_seatemp.t, unit='s')
  if subset:
    df_seatemp = df_seatemp[(df_seatemp.new_timestamp <=  pd.to_datetime('2018 2 16')) & (df_seatemp.new_timestamp >=  pd.to_datetime('2017 11 1'))]
  else:
    df_seatemp = df_seatemp[(df_seatemp.new_timestamp <=  pd.to_datetime('2019 4 3')) & (df_seatemp.new_timestamp >=  pd.to_datetime('2017 11 1')) ]
  df_seatemp = df_seatemp[['new_timestamp', 't', 'v']]
  return df_seatemp.reset_index(drop=True)

def join_temp_curr_seatemp(df_segment, df_current, df_seatemp):
  df_current = df_current.set_index('t')
  df_joined =  df_segment[[segment, 'timestamp']].set_index('timestamp').join(df_current['v'], how='outer')
  df_joined = df_joined.join(df_seatemp[['v', 't']].set_index('t'), how='outer', rsuffix='_temp')
  df_joined['v'] = df_joined['v'].interpolate()
  df_joined['v_temp'] = df_joined['v_temp'].interpolate()
  df_joined = df_joined.dropna().reset_index()

  df_joined.columns = ["timestamp", df_joined.columns[1], 'current', 'sea_temp']
  df_joined['sampling_int'] = df_joined.timestamp - df_joined.timestamp.shift(1)
  df_joined.loc[0, 'sampling_int'] = 0
  df_joined['new_timestamp'] = pd.to_datetime(df_joined.timestamp, unit='s')
  return df_joined

# COMMAND ----------

def make_time_ranges(start_time, end_time, duration):
  range_start_time = start_time
  range_end_time = start_time + duration
  time_ranges = []
  while range_end_time < end_time:
    time_ranges.append({'start_timestamp': range_start_time, 'end_timestamp':range_end_time})
    range_start_time = range_end_time
    range_end_time = range_start_time + duration
  return time_ranges

def df_by_time_range(df, time_range):
  return df.loc[(df.new_timestamp >= time_range['start_timestamp']) & (df.new_timestamp < time_range['end_timestamp']),:]

def get_time_ranges():
  return make_time_ranges(pd.Timestamp('2017-11-01 00:00:00'), pd.Timestamp('2019-4-03 00:00:00'), pd.Timedelta('2 days'))

def get_time_ranges_subset():
  return make_time_ranges(pd.Timestamp('2017-11-01 00:00:00'), pd.Timestamp('2018-2-16 00:00:00'), pd.Timedelta('2 days'))

def get_fold_ranges(subset=False, n_folds=n_folds, random_seed=None):
  if subset:
    time_ranges = get_time_ranges_subset()
  else:
    time_ranges = get_time_ranges()
  if random_seed:
    print("Warning: random seed changed from default (20)")
  else:
    random_seed = 20
  n_blocks = int(np.ceil(len(time_ranges)/n_folds))
  fold_indecies = [range(n*n_blocks,(n+1)*n_blocks) for n in range(n_folds)]
  random.seed(random_seed)
  shuffled_ranges = time_ranges.copy()
  random.shuffle(shuffled_ranges)

  return [shuffled_ranges[fold_index.start:fold_index.stop] for fold_index in fold_indecies]

# COMMAND ----------

def set_dirs(name):
  name = name if name.endswith('/') else name +'/'
  
  log_dir = '/dbfs/mnt/renewablesstudents/user-write/logfiles/p2/' + name 
  model_dir = '/dbfs/mnt/renewablesstudents/user-write/models/p2/' + name 
  
  if not os.path.exists(log_dir):
      os.makedirs(log_dir)
      print('dir made: {}'.format(log_dir))
      
  if not os.path.exists(model_dir):
      os.makedirs(model_dir)
      print('dir made: {}'.format(model_dir))
  
  return log_dir, model_dir

# COMMAND ----------

class Data_by_fold():
  def __init__(self, test_fold, val_fold, fold_ranges, curr=False, st=False, dts=False):
    assert test_fold != val_fold, 'test fold cannot equal val fold.'
    self.test_ranges = fold_ranges[test_fold]
    self.val_ranges = fold_ranges[val_fold]
    self.train_ranges = [fold_ranges[fold][j] for fold in range(n_folds) if fold != val_fold and fold != test_fold for j in range(len(fold_ranges[fold]))]
    
    if curr:
      self.curr_mean, self.curr_std = self._get_mean_std([df_by_time_range(df_current, tr).v for tr in self.train_ranges])
    if st:
      self.st_mean, self.st_std = self._get_mean_std([df_by_time_range(df_seatemp, tr).v for tr in self.train_ranges])
    if dts:
      self.dts_mean, self.dts_std = self._get_mean_std([df_by_time_range(df_dts, tr).iloc[:,:n_segments] for tr in self.train_ranges], axis=0)
    
  def _get_mean_std(self, data, axis=None):
    data = np.concatenate(data)
    return np.mean(data, axis=axis), np.std(data, axis=axis)
  
  def get_training_data(self):
    raise NotImplementedError
    
  def get_validation_data(self):
    raise NotImplementedError
    
  def get_anomaly_test_data(self):
    raise NotImplementedError
    # Load generated anomalies from drive and metadata.
    # Also df with no anomaly!

# COMMAND ----------

class Log:
  def __init__(self, log_dir=False, file_suffix=False):
    self.is_logging = False
    if file_suffix:
      self.log_path = log_dir + file_suffix + '.txt'
      self.is_logging = True
      assert not os.path.isfile(self.log_path), "path " + self.log_path + " is already in use"
      self.write("Logging to {}".format(self.log_path))
    else:
      print("Session is not logged")
    
  def write(self, s):
    if self.is_logging:
      with open(self.log_path, 'a') as file:
        file.write('{}: '.format(datetime.now()) + s + '\n')
    print(s)

# COMMAND ----------

def get_fault_data(data_path, metadata_path):
  df_loaded = pd.read_csv(data_path)
  with open(metadata_path) as file:
    metadata = json.load(file)
  assert set(metadata['segments']).issubset(set(list(df_loaded))), 'segments in metadata does not match segments in data'
  return df_loaded, metadata

# COMMAND ----------

def save_params(sess, epoch, test_fold, val_fold, log, model_dir):
  if model_dir:
    epoch_dir = model_dir + '/params_t{}v{}/'.format(test_fold, val_fold, epoch)
    if os.path.exists(epoch_dir):
      shutil.rmtree(epoch_dir)
    os.makedirs(epoch_dir)
    fp = saver.save(sess, epoch_dir + 'model.ckpt')
    log.write('\tparams saved to {}'.format(fp))
    return fp
  else: return None

# COMMAND ----------

class AnomalyDetector:
  """
  AD model.
  """
  def __init__(self, model_dir=None):
    pass
  
  def is_anomalous(self, data, test_fold=-1):
    raise NotImplementedError
    
class Random_ad(AnomalyDetector):
  def __init__(self):
    pass
  
  def is_anomalous(self, data):
    return random.choice([True,False])

# COMMAND ----------

def model_summary(log):
    log.write('---------')
    log.write('Variables: name (type shape) [size]')
    log.write('---------')
    model_vars = tf.trainable_variables()
    for var in model_vars:
      var_size = var.get_shape().num_elements() or 0
      var_bytes = var_size * var.dtype.size
      log.write("{}: {}, [{}, bytes: {}]".format(var.name, tf.contrib.slim.model_analyzer.tensor_description(var), var_size, var_bytes))
    log.write('---------\n')

# COMMAND ----------

'''
def get_result_summary(result_list, metadata):
  raise ValueError("USE get_result_summary_V2 and get_AD_results_V2")
  if metadata['type'] == 'instant':
    columns = metadata['columns']
    begin_step = metadata['begin_step']
    result_list[begin_step:]
    located_anomaly= [result[0] in columns and result[0]>-1  for result in result_list[begin_step:]]
    proportion_found = np.mean(located_anomaly) if located_anomaly else 0
    is_discovered = np.any(located_anomaly)
    if True in located_anomaly:
      delay = located_anomaly.index(True)
    else:
      delay = -999999
    pct_increase = metadata['pct_increase']
  elif metadata['type'] == 'drifting':
    columns = metadata['columns']
    located_anomaly = [result[0] in columns and result[0]>-1 for result in result_list]
    proportion_found = np.mean(located_anomaly) if located_anomaly else 0
    is_discovered = np.any(located_anomaly)
    if True in located_anomaly:
      delay = located_anomaly.index(True)
    else:
      delay = -999999
    pct_increase = metadata['increase_list'][delay] if delay >= 0 else -1
  elif metadata['type'] == 'no_fault':
    located_anomaly = [result[0] == -1 for result in result_list]
    proportion_found = np.mean(located_anomaly) if located_anomaly else 0
    is_discovered =  np.all(located_anomaly)
    delay=-1
    pct_increase=-1
  else:
    print(metadata['type'])
    raise ValueError
  return proportion_found, delay, pct_increase, is_discovered

def get_AD_results(anomalyDetector, fault_dir, log=False, column='random'):
  raise ValueError("USE get_AD_results_V2 in          func AD evaluation")
  if not log:
    log = Log(False, False)
  results = []
  log.write('Finding anomaly detection results for column {}'.format(column))
  types = ['typeI', 'typeII', 'typeIII', 'typeIV']
  
  for fold in range(n_folds):
    fold_results = {}
    log.write('fold: {}'.format(fold))
    fold_dir = fault_dir + 'test_fold{}/'.format(fold) + 'column_{}/'.format(column)  
    for fault_type in types:
      fold_results[fault_type] = {}
      type_results = []
      data_paths = glob.glob(fold_dir + fault_type + '/*.csv')
      metadata_paths = glob.glob(fold_dir + fault_type + '/*.json')
      assert len(data_paths) > 0, 'data paths is empty: {}'.format(fold_dir + fault_type + '/*.csv')
      for index, data_path in enumerate(data_paths):
        metadata_path = metadata_paths[index]
        try:
          fault_df, metadata = get_fault_data(data_path, metadata_path)
        except AssertionError:
          print("Is only subset loaded? Faults stem from all data!")
          raise
        type_results.append(get_result_summary(anomalyDetector.is_anomalous(fault_df, metadata, fold), metadata))
      fold_results[fault_type]['all'] = type_results
      fold_results[fault_type]['aggregated'] = {'avg proportion': np.mean([ts[0] for ts in type_results]), 'avg delay': np.mean([ts[1] for ts in type_results if ts[1] >= 0]), 'proportion discovered': np.mean([ts[3] for ts in type_results])}
      log.write('\t' + fault_type + ': ' + str(fold_results[fault_type]['aggregated']))

    # Fault-free samples:
    type_results = []
    fault_type = 'no_fault'
    fold_results[fault_type] = {}
    for tr in fold_ranges[fold]:
      fault_df = df_by_time_range(df_segment, tr).copy()
      metadata = {'type':'no_fault', 'start':tr['start_timestamp'], 'end':tr['end_timestamp']}
      type_results.append(get_result_summary(anomalyDetector.is_anomalous(fault_df, metadata, fold), metadata))
    fold_results[fault_type]['all'] = type_results
    fold_results[fault_type]['aggregated'] = {'avg proportion': np.mean([ts[0] for ts in type_results]), 'avg delay': np.mean([ts[1] for ts in type_results if ts[1] >= 0]), 'proportion discovered': np.mean([ts[3] for ts in type_results])}
    log.write('\t' + fault_type + ': ' + str(fold_results[fault_type]['aggregated']))
    results.append(fold_results)
    
  log.write('Aggregated results:')
  
  all_faults = {ft:[] for ft in ['typeI', 'typeII', 'typeIII', 'typeIV', 'no_fault']}
  for fold in range(n_folds):
    for fault_type in ['typeI', 'typeII', 'typeIII', 'typeIV', 'no_fault']:
      all_faults[fault_type].extend(results[fold][fault_type]['all'])
  for fault_type, fault_list in all_faults.items():
    log.write(fault_type + ': ' + str({'avg proportion': np.mean([ts[0] for ts in fault_list]), 'avg delay': np.mean([ts[1] for ts in fault_list if ts[1] >= 0]), 'proportion discovered': np.mean([ts[3] for ts in fault_list])}))

  return results
'''
pass

# COMMAND ----------

def get_reconstructions(sess, data):
  recs = []
  for batch in range(0, len(data), batch_size):
    rec = sess.run(r, feed_dict={x:data[batch:batch+batch_size], bs:len(data[batch:batch+batch_size])})
    if isinstance(rec, list):
      recs.append(np.swapaxes(np.array(rec), 0, 1))
    else:
      recs.append(rec)
  return np.concatenate(recs)

def get_ae_loss(sess, data):
  recs = get_reconstructions(sess, data)
  losses = np.square(data-recs)
  return np.mean(losses)

def get_recs_from_saved_params(params_path, data):
  with tf.Session() as sess:
    saver.restore(sess, params_path)
    return get_reconstructions(sess, data)

def get_rec_error_from_saved_params(params_path, data):
  return np.reshape(data - get_recs_from_saved_params(params_path, data), (len(data), -1))

def train_AE_network(data_by_fold, n_epochs=10, batch_size=128, initial_patience=50, file_suffix=False, all_test_folds=True, log=False, print_loss_op=True):
  if not file_suffix:
    print("MODELS ARE NOT SAVED")
  if not log:
    log = Log(False, False)
  model_dir = models_dir + str(file_suffix)
  model_summary(log)

  for test_fold in range(n_folds if all_test_folds else 1):
    val_fold = (test_fold+1) % n_folds
    if val_fold != test_fold:
      log.write('test fold: {}, val fold: {}'.format(test_fold, val_fold))
      patience = initial_patience
      sd = data_by_fold(test_fold, val_fold, fold_ranges)
      train_data = sd.get_training_data()
      val_data = sd.get_validation_data()      
      min_val_loss = 1
      epoch_inputs = train_data
      with tf.Session() as sess:
        sess.run(tf.global_variables_initializer())
        log.write('Prior val loss {}'.format(get_ae_loss(sess, val_data)))
        for epoch in range(n_epochs):
          np.random.shuffle(epoch_inputs)
          for batch in range(0, len(epoch_inputs), batch_size):
            batch_inputs = epoch_inputs[batch: batch+batch_size, :]
            sess.run(train_op, feed_dict={x:batch_inputs, bs:len(batch_inputs)})
          epoch_loss = get_ae_loss(sess, train_data)
          epoch_val_loss = get_ae_loss(sess, val_data)
          if print_loss_op:
            log.write('epoch {} finished with train loss {}, val loss {}. Loss op: {}'.format(epoch, epoch_loss, epoch_val_loss, sess.run(loss_op, feed_dict={x:train_data, bs:len(train_data)})))
          else:
            log.write('epoch {} finished with train loss {}, val loss {}'.format(epoch, epoch_loss, epoch_val_loss))
          if epoch_val_loss < min_val_loss:
            min_val_loss = epoch_val_loss
            patience = initial_patience
            if file_suffix:
              params_path = save_params(sess, epoch, test_fold, val_fold, log, model_dir)
            else:
              print('calculated val_recs')
              val_recs = get_reconstructions(sess, val_data)
          else:
            patience -= 1
            if patience == 0:
              log.write('Breaking due to no val improvement in {} epochs'.format(initial_patience))
              break
  log.write("Done!")
  if not file_suffix:
    return val_recs
  else:
    return None

# COMMAND ----------


