# Databricks notebook source
# MAGIC %run ../spatial/func_spatial

# COMMAND ----------

# MAGIC %run ../temporal/func_temporal

# COMMAND ----------

class Spatiotemporal_data_by_fold(Data_by_fold):
  def __init__(self, test_fold, val_fold, fold_ranges, pca_components=False):
    super().__init__(test_fold, val_fold, fold_ranges, curr=True, st=True, dts=True)
    self.interval_mean, self.interval_std = self._get_mean_std([df_by_time_range(df_dts_joined, tr)['sampling_int'] for tr in self.train_ranges])
    if pca_components:
      self.PCA = PCA(pca_components, svd_solver='auto')
      self.pca_components = pca_components
      self.encoding_mean, self.encoding_std = self.__fit_PCA()
    else:
      self.PCA = False
      self.pca_components = False
  
  def __normalize(self, data, mean, std):
    return (data - mean) / std
  
  def __get_inputs_targets(self, data, interval_list, not_PCA):
    out_dim = n_segments
    if self.PCA and not not_PCA:
      pca_seg = self.__normalize(self.PCA.transform(data[:,:n_segments]), self.encoding_mean, self.encoding_std)
      data = np.concatenate([pca_seg, data[:,n_segments:]], axis=1)
      out_dim =  self.pca_components
    size = len(data) - INPUT_STEPS - OUTPUT_STEPS + 1
    dim = data.shape[1]
    inputs = np.zeros((size, INPUT_STEPS, dim))
    targets = np.zeros((size, OUTPUT_STEPS, out_dim))
    intervals = np.zeros((size, OUTPUT_STEPS))
    for i in range(size):
      inputs[i] = data[i:i+INPUT_STEPS,:]
      targets[i] = data[i+INPUT_STEPS:i+INPUT_STEPS+OUTPUT_STEPS, :out_dim]
      intervals[i] = interval_list[i+INPUT_STEPS:i+INPUT_STEPS+OUTPUT_STEPS]
    return inputs, targets.reshape((size, -1)), intervals
  
  def __get_input_data_by_ranges(self, ranges, not_PCA=False):
    inputs = []
    targets = []
    intervals = []
    for tr in ranges:
      range_df = df_by_time_range(df_dts_joined, tr).drop(['timestamp', 'new_timestamp'], axis=1)
      range_df.iloc[:,:n_segments] = self.__normalize(range_df.iloc[:,:n_segments].values, self.dts_mean, self.dts_std)
      range_df.loc[:,'current'] = self.__normalize(range_df['current'].values, self.curr_mean, self.curr_std)
      range_df.loc[:,'sea_temp'] = self.__normalize(range_df['sea_temp'].values, self.st_mean, self.st_std)
      range_df.loc[:,'sampling_int'] = self.__normalize(range_df['sampling_int'].values, self.interval_mean, self.interval_std)
      range_inputs, range_targets, range_intervals = self.__get_inputs_targets(range_df.values, range_df['sampling_int'].values, not_PCA)
      inputs.append(range_inputs)
      targets.append(range_targets)
      intervals.append(range_intervals)
    return np.concatenate(inputs), np.concatenate(targets), np.concatenate(intervals)
  
  def __fit_PCA(self):
    segment_dfs = []
    for tr in self.train_ranges:
      range_df = df_by_time_range(df_dts_joined, tr).drop(['timestamp', 'new_timestamp'], axis=1)
      segment_dfs.append(self.__normalize(range_df.iloc[:,:n_segments].values, self.dts_mean, self.dts_std))
    transformed = self.PCA.fit_transform(np.concatenate(segment_dfs))
    print('PCA fit with {} components explaining {} of the variance'.format(self.PCA.n_components_, np.sum(self.PCA.explained_variance_ratio_)))
    encoding_mean = np.mean(transformed, axis=0)
    encoding_std = np.std(transformed, axis=0)
    return encoding_mean, encoding_std
  
  def get_training_data(self):
    return self.__get_input_data_by_ranges(self.train_ranges)
    
  def get_validation_data(self):
    return self.__get_input_data_by_ranges(self.val_ranges)

  def get_inputs_targets_intervals_from_df(self, df_input):
    df = df_input.drop(['timestamp', 'new_timestamp'], axis=1)
    
    df.iloc[:,:n_segments] = self.__normalize(df.iloc[:,:n_segments].values, self.dts_mean, self.dts_std)
    df.loc[:,'current'] = self.__normalize(df.loc[:,'current'].values, self.curr_mean, self.curr_std)
    df.loc[:,'sea_temp'] = self.__normalize(df.loc[:,'sea_temp'].values, self.st_mean, self.st_std)
    df.loc[:,'sampling_int'] = self.__normalize(df.loc[:,'sampling_int'].values, self.interval_mean, self.interval_std)
    
    pca_inputs, _, intervals = self.__get_inputs_targets(df.values, df['sampling_int'].values, not_PCA=False)
    _, dts_targets, _ = self.__get_inputs_targets(df.values, df['sampling_int'].values, not_PCA=True)
    return pca_inputs, dts_targets, intervals

  def get_training_data_not_PCA(self):
    return self.__get_input_data_by_ranges(self.train_ranges, not_PCA=True)
  
  def get_validation_data_not_PCA(self):
    return self.__get_input_data_by_ranges(self.val_ranges, not_PCA=True)
  
class Spatiotemporal_data_by_fold_PCA(Spatiotemporal_data_by_fold):
  def __init__(self, test_fold, val_fold, fold_ranges):
    super().__init__(test_fold, val_fold, fold_ranges, pca_components=PCA_COMPONENTS)
    
  def get_sequence_from_pca(self, pca_data):
    unnormalized_pca_data = pca_data * self.encoding_std + self.encoding_mean
    seg = unnormalized_pca_data @ self.PCA.components_
    return seg

# COMMAND ----------

def get_predictions(sess, inputs, intervals, BS=False, batch_size=batch_size):
  predictions = []
  for batch in range(0, len(inputs), batch_size):
    if BS:
      predictions.append(sess.run(p, feed_dict={x:inputs[batch:batch+batch_size], t:intervals[batch:batch+batch_size], bs:len(inputs[batch:batch+batch_size])}))
    else:
      predictions.append(sess.run(p, feed_dict={x:inputs[batch:batch+batch_size], t:intervals[batch:batch+batch_size]}))
  return np.concatenate(predictions)

def get_prediction_loss(sess, inputs, targets, intervals, BS=False, batch_size=batch_size, loss_weights=False):
  predictions = get_predictions(sess, inputs, intervals, BS, batch_size)
  losses = np.square(targets - predictions)
  if loss_weights:
    losses = losses * loss_weights
  return np.mean(losses)

def get_prediction_loss_PCA(sess, PCA_inputs,  DTS_targets, intervals,  dbf):
  predictions = get_predictions(sess, PCA_inputs, intervals)
  DTS_preds = dbf.get_sequence_from_pca(predictions)
  losses = np.square(DTS_targets - DTS_preds)
  return np.mean(losses)

# COMMAND ----------

def ST_model(lr=0.001, log=False, n_hidden=500):
  dim_input = PCA_COMPONENTS+3
  dim_output = PCA_COMPONENTS
  
  if not log:
    log = Log(False, False)
  tf.reset_default_graph()
  
  params = {'learning rate': lr, 'LSTM hidden units':n_hidden, 'dim input': dim_input, 'dim output':dim_output, 'input steps': INPUT_STEPS, 'output steps': OUTPUT_STEPS}
  log.write(str(params))
  
  # Graph input
  x = tf.placeholder("float", [None, INPUT_STEPS,  dim_input], name='x')  #input
  y = tf.placeholder("float", [None, OUTPUT_STEPS*dim_output], name='y')  #target
  t = tf.placeholder("float", [None, OUTPUT_STEPS], name='t')  # time
  w = tf.placeholder("float")
  
  weights_out = tf.Variable(tf.random_normal([n_hidden+OUTPUT_STEPS, OUTPUT_STEPS*dim_output]))   # +n_output_steps because of concatenated time until prediction input
  biases_out = tf.Variable(tf.zeros([OUTPUT_STEPS*dim_output]))

  cell = rnn_cell.DropoutWrapper(rnn_cell.LSTMCell(n_hidden, name='LSTM_cell'), output_keep_prob=1)
  outputs, _ = tf.nn.dynamic_rnn(cell, x, dtype=tf.float32)  
  p = tf.matmul(tf.concat([outputs[:, -1, :], t], axis=1), weights_out) + biases_out
  p = tf.identity(p, name='p')
  
  loss_op = tf.losses.mean_squared_error(labels=y, predictions=p, weights=w)
  optimizer = tf.train.AdamOptimizer(learning_rate=lr)
  train_op = optimizer.minimize(loss_op)
  
  return train_op, p, x, t, y, w

# COMMAND ----------

def trainPCA_PredictionNetwork(n_epochs=10, batch_size=128, initial_patience=50, file_suffix=False, all_test_folds=True, log=False,
                               dbf=Spatiotemporal_data_by_fold_PCA, save_model=True, use_loss_weights=False):
  if not log:
    log = Log()
  model_dir = models_dir + str(file_suffix)
  model_summary(log)
  
  params = {'save_model':save_model, 'n_epochs':n_epochs, 'batch_size':batch_size, 'initial_patience':initial_patience,
            'file_suffix':file_suffix, 'all_test_folds':all_test_folds,  'use_loss_weights':use_loss_weights}
  log.write(str(params))
  
  for test_fold in range(n_folds if all_test_folds else 1):
    val_fold = (test_fold+1) % n_folds
    if val_fold != test_fold:
      log.write('test fold: {}, val fold: {}'.format(test_fold, val_fold))
      val_predictions = False
      patience = initial_patience
      td = dbf(test_fold, val_fold, fold_ranges)
      train_inputs, train_targets, train_intervals = td.get_training_data()
      val_inputs, val_targets, val_intervals = td.get_validation_data() 
      _, train_PCA_targets, _ = td.get_training_data_not_PCA()
      _, val_PCA_targets, _ = td.get_validation_data_not_PCA()
      loss_weights = td.PCA.explained_variance_ratio_.reshape(1, -1).tolist() if use_loss_weights else 1
      if test_fold == 0: print("val inputs, targets, intervals:", val_inputs.shape, val_targets.shape, val_intervals.shape)
      epoch_indicies = [i for i in range(len(train_inputs))]
      with tf.Session() as sess:
        sess.run(tf.global_variables_initializer())
        epoch_val_loss = get_prediction_loss(sess, val_inputs, val_targets, val_intervals, loss_weights=loss_weights)
        epoch_val_PCA_loss = get_prediction_loss_PCA(sess, val_inputs, val_PCA_targets, val_intervals, dbf=td)
        min_val_loss = epoch_val_PCA_loss if use_loss_weights else epoch_val_loss
        log.write('Prior losses' )
        epoch_train_loss = get_prediction_loss(sess, train_inputs, train_targets, train_intervals, loss_weights=loss_weights)
        epoch_train_PCA_loss = get_prediction_loss_PCA(sess, train_inputs, train_PCA_targets, train_intervals, dbf=td)
        log.write('epoch {} finished with train loss {} , val loss {} . For whole sequence train loss was {} and val loss was {}'.format(-1, epoch_train_loss, epoch_val_loss,
                                                                                                                                         epoch_train_PCA_loss, epoch_val_PCA_loss))
        if use_loss_weights:
            epoch_mean_train_loss = get_prediction_loss(sess, train_inputs, train_targets, train_intervals)
            epoch_mean_val_loss = get_prediction_loss(sess, val_inputs, val_targets, val_intervals)
            log.write('\t PCA MSE: t: {} v: {}'.format(epoch_mean_train_loss, epoch_mean_val_loss))
        for epoch in range(n_epochs):
          np.random.shuffle(epoch_indicies)
          epoch_inputs = train_inputs[epoch_indicies]
          epoch_targets = train_targets[epoch_indicies]
          epoch_intervals = train_intervals[epoch_indicies]
          for batch in range(0, len(epoch_indicies), batch_size):
            batch_inputs = epoch_inputs[batch: batch+batch_size, :]
            batch_targets = epoch_targets[batch: batch+batch_size, :]
            batch_intervals = epoch_intervals[batch: batch+batch_size, :]
            sess.run(train_op, feed_dict={x:batch_inputs, y:batch_targets, t:batch_intervals, w:loss_weights})
          epoch_train_loss = get_prediction_loss(sess, train_inputs, train_targets, train_intervals, loss_weights=loss_weights)
          epoch_val_loss =  get_prediction_loss(sess, val_inputs, val_targets, val_intervals, loss_weights=loss_weights)
          epoch_train_PCA_loss = get_prediction_loss_PCA(sess, train_inputs, train_PCA_targets, train_intervals, dbf=td)
          epoch_val_PCA_loss =  get_prediction_loss_PCA(sess, val_inputs, val_PCA_targets, val_intervals, dbf=td)
          log.write('epoch {} finished with train loss {} , val loss {} . For whole sequence train loss was {} and val loss was {}'.format(epoch, epoch_train_loss, epoch_val_loss,
                                                                                                                                           epoch_train_PCA_loss, epoch_val_PCA_loss))
          if use_loss_weights:
            epoch_mean_train_loss = get_prediction_loss(sess, train_inputs, train_targets, train_intervals)
            epoch_mean_val_loss = get_prediction_loss(sess, val_inputs, val_targets, val_intervals)
            log.write('\t PCA MSE: t: {} v: {}'.format(epoch_mean_train_loss, epoch_mean_val_loss))
          if (epoch_val_PCA_loss if use_loss_weights else epoch_val_loss) < min_val_loss:
            min_val_loss = epoch_val_PCA_loss if use_loss_weights else epoch_val_loss
            patience = initial_patience
            if file_suffix and save_model:
              params_path = save_params(sess, epoch, test_fold, val_fold, log, model_dir)
            else:
              val_predictions = get_predictions(sess, val_inputs, val_intervals)
              train_predictions = get_predictions(sess, train_inputs, train_intervals)
              print('preds calculated')
          else:
            patience -= 1
            if patience == 0:
              log.write('Breaking due to no val improvement in {} epochs'.format(initial_patience))
              print('val_predictions', val_predictions)
              if not val_predictions:
                val_predictions = get_predictions(sess, val_inputs, val_intervals)
                train_predictions = get_predictions(sess, train_inputs, train_intervals)
                print('preds calculated')
              break
  log.write("Done!")
  if file_suffix and save_model:
    with tf.Session() as sess:
      saver.restore(sess, params_path)
      val_preds = get_predictions(sess, val_inputs, val_intervals)
      train_preds = get_predictions(sess, train_inputs, train_intervals)
  return val_preds, train_preds

# COMMAND ----------

def join_dts_curr_seatemp(df_dts, df_current, df_seatemp):
  df_current = df_current.set_index('t')
  df_joined =  df_dts.drop(['new_timestamp', 'sampling_interval'], axis=1).set_index('timestamp').join(df_current['v'], how='outer')
  df_joined = df_joined.join(df_seatemp[['v', 't']].set_index('t'), how='outer', rsuffix='_temp')
  df_joined['v'] = df_joined['v'].interpolate()
  df_joined['v_temp'] = df_joined['v_temp'].interpolate()
  df_joined = df_joined.dropna().reset_index()
  
  cols = df_joined.columns.values
  cols[0] = "timestamp"
  cols[-2] = 'current' 
  cols[-1] = 'sea_temp'
  
  df_joined['sampling_int'] = df_joined.timestamp - df_joined.timestamp.shift(1)
  df_joined.loc[0, 'sampling_int'] = 0
  df_joined['new_timestamp'] = pd.to_datetime(df_joined.timestamp, unit='s')
  return df_joined

# COMMAND ----------



# COMMAND ----------

'''
def ST_model(lr=0.001, log=False):
  lstm_n_hidden = 64
  dim_input = n_segments+3
  
  if not log:
    log = Log(False, False)
  tf.reset_default_graph()
  
  params = {'learning rate': lr, 'LSTM hidden units':lstm_n_hidden, 'dim input': dim_input, 'input steps': INPUT_STEPS, 'output steps': OUTPUT_STEPS}
  log.write(str(params))
  
  # Graph input
  x = tf.placeholder("float", [None, INPUT_STEPS,  dim_input], name='x')  #input
  y = tf.placeholder("float", [None, OUTPUT_STEPS*n_segments], name='y')  #target
  t = tf.placeholder("float", [None, OUTPUT_STEPS], name='t')  # time
  bs = tf.placeholder("int32", name='bs') # batch size
  
  # CAE params
  n_filters = 30
  ksize1 = [5, 1, n_filters]
  ksize2 = [5, n_filters, n_filters]
  ksize3 = [5, n_filters, n_filters]
  ksize4 = [5, n_filters, 4]
  p1 = 6
  p2 = 6
  p3 = 6
  p4 = 5
  
  log.write('ksize 1: {}, max pool 1: {}'.format(ksize1, p1))
  log.write('ksize 2: {}, max pool 2: {}'.format(ksize2, p2))
  log.write('ksize 3: {}, max pool 3: {}'.format(ksize3, p3))
  log.write('ksize 4: {}, max pool 4: {}'.format(ksize4, p4))

  W = tf.get_variable('W', shape=(n_filters,1), initializer=tf.random_normal_initializer())
  B = tf.get_variable('B', shape=(1), initializer=tf.zeros_initializer())
  


  segment_inp = tf.reshape(x[:,:,:n_segments], shape=[-1, n_segments, 1],name="reshape_seg_input")  
  aux_input = x[:,:,n_segments:]
  
  print('segment_inp', segment_inp.shape, 'aux_input', aux_input.shape )
  log.write('input {}'.format(segment_inp))

  c1, s1 = conv(segment_inp, ksize1, 1, 'conv1')
  log.write('conv1: {}'.format(c1))
  mp1 = max_pool(c1, p1, p1, 'mp1', ksize1[2], bs*INPUT_STEPS)
  log.write('mp1: {}'.format(mp1))
  c2, s2 = conv(mp1, ksize2, 1, 'conv2')
  log.write('conv2: {}'.format(c2))
  mp2 = max_pool(c2, p2, p2, 'mp2', ksize2[2], bs*INPUT_STEPS)
  log.write('mp2: {}'.format(mp2))
  c3, s3 = conv(mp2, ksize3, 1, 'conv3')
  log.write('c3: {}'.format(c3))
  mp3 = max_pool(c3, p3, p3, 'mp3', ksize3[2], bs*INPUT_STEPS)
  log.write('mp3: {}'.format(mp3))
  c4, s4 = conv(mp3, ksize4, 1, 'conv4')
  log.write('c4: {}'.format(c4))
  mp4 = max_pool(c4, p4, p4, 'mp4', ksize4[2], bs*INPUT_STEPS)
  log.write('mp4: {}'.format(mp4))
  
  encoding = tf.identity(mp4, name='encoding')
  encoding_size = ksize4[2] * int(np.ceil(n_segments/(p1*p2*p3*p4)))
  encoding = tf.reshape(encoding, shape=[bs, INPUT_STEPS, encoding_size], name='encoding_reshape')
  log.write('\t encoding, {}'.format(encoding))
  
  
  # LSTM
  LSTM_weight = tf.Variable(tf.random_normal([lstm_n_hidden+OUTPUT_STEPS, OUTPUT_STEPS*encoding_size]))   # +n_output_steps because of concatenated time until prediction input
  LSTM_bias = tf.Variable(tf.zeros([OUTPUT_STEPS*encoding_size]))
  cell = rnn_cell.LSTMCell(lstm_n_hidden, name='LSTM_cell')

  LSTM_input = tf.concat([encoding, aux_input], axis=2)
  log.write('\tLSTM input {}'.format(LSTM_input.shape))
  
  outputs, _ = tf.nn.dynamic_rnn(cell, LSTM_input, dtype=tf.float32)  
  log.write('\tLSTM outputs {}'.format(outputs))
  
  p = tf.matmul(tf.concat([outputs[:, -1, :], t], axis=1), LSTM_weight) + LSTM_bias
  lstm_prediction = tf.identity(p, name='p')
  log.write('LSTM prediction {}'.format(lstm_prediction))
  
  lstm_prediction = tf.reshape(lstm_prediction, shape=[bs, -1, ksize4[2]])  
  
  dc4 = deconv(lstm_prediction, ksize4, s4, 'deconv4', int(np.ceil(n_segments/(p1*p2*p3*p4))), bs)
  log.write('dc4: {}'.format(dc4))
  up4 = upsample(dc4, 'unpool4', int(np.ceil(n_segments/(p1*p2*p3))), ksize4[1], bs)
  log.write('up4: {}'.format(up4))
  dc3 = deconv(up4, ksize3, s3,'deconv3', int(np.ceil(n_segments/(p1*p2*p3))), bs)
  log.write('dc3: {}'.format(dc3))
  up3 = upsample(dc3, 'unpool0', int(np.ceil(n_segments/(p1*p2))), ksize3[1], bs)
  log.write('up3: {}'.format(up3))
  dc2 = deconv(up3, ksize2, s2,'deconv2', int(np.ceil(n_segments/(p1*p2))), bs)
  log.write('dc2: {}'.format(dc2))
  up2 = upsample(dc2, 'unpool1', n_segments//p1, ksize2[1], bs)
  log.write('up2: {}'.format(up2))
  dc1 = deconv(up2, [ksize1[0],  ksize2[1], n_filters], s1,'deconv1', int(np.ceil(n_segments//p1)), bs)
  log.write('dc1: {}'.format(dc1))
  up1 = upsample(dc1, 'unpool1', n_segments, ksize1[2], bs)
  log.write('up1: {}'.format(up1))

  res = tf.matmul(tf.reshape(up1, shape = [-1, n_filters]), W) + B
  log.write('res: {}'.format(res))

  r = tf.reshape(res, shape=[-1, n_segments])

  loss_op = tf.losses.mean_squared_error(labels=y, predictions=r)
  optimizer = tf.train.AdamOptimizer(learning_rate=lr)
  train_op = optimizer.minimize(loss_op)
  
  return train_op, r, x, t, y, bs
train_op, p, x, t, y, bs = ST_model()'''
pass
